// Copyright (C) 2008 David Bremner
//
// Author: David Bremner (bremner@unb.ca)
// Version: $Id$
//
// This file is part of lpmargin.
//
// lpmargin is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// lpmargin is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with lpmargin.  If not, see <http://www.gnu.org/licenses/>.
#ifndef _DB_MATRIX_H
#define _DB_MATRIX_H
#include <vector>

#include "Vector.h"

template <class T>
class Matrix  {
 private:
  std::vector < Vector<T> > rowlist;
 public:
  typedef Vector<T>  vec;
  typedef std::vector< vec > vecvec;
  typedef Matrix<T> mat;
  typedef typename vecvec::const_iterator const_iterator;
  typedef  T element_t;

  // Constructors
  Matrix() {};
  Matrix(int r, int c) : rowlist(r,vec(c)) {};
  Matrix(const mat & M) {
    const_iterator j;

    for (j=M.begin(); j!=M.end(); j++){
      append(*j);
    }
  }

  static mat Identity(int d){
    mat I(d,d);
    for (int i=0; i<d; i++){
      I[i][i]=static_cast<T>(1);
    }
    return I;
  }


  int rows() const { return rowlist.size(); }
  int cols() const { return rowlist[0].size(); }

  // iterators
  typename vecvec::iterator begin() { return rowlist.begin(); };
  
  typename vecvec::const_iterator begin() const { return rowlist.begin(); };
  typename vecvec::iterator end() { return rowlist.end(); };
  typename vecvec::const_iterator end() const { return rowlist.end(); };


  vec &operator [] (int i) { return rowlist[i]; };
  const vec &operator [] (int i) const { return rowlist[i]; };

  void append (const vec& v){
    rowlist.push_back(v);
  }

  void append(const mat& m) { 
    typename vecvec::const_iterator j; 
    for (j=m.begin(); j!=m.end(); j++)
      append(*j);
  };

  void append_col(const vec& v) { 
    for (int j=0; j<rows(); j++)
      rowlist[j].append(v[j]);
  };

  mat  transpose() const{
    
    mat M(cols(),rows());
    for (int i=0; i< rows(); i++)
      for (int j=0; j<cols(); j++){
	M[j][i]=(*this)[i][j];
      }
    return M;
  }
      
  void mul(const T& scalar){
    typename vecvec::iterator j;
    for(j=rowlist.begin(); j!=rowlist.end(); j++){
      j->mul(scalar);
    }
  }

  int dim(){
    mat W(*this);


    vec v(rows(),1);


    W.append_col(v);
    

    return W.destrank()-1;
  }
  int rank(){
    mat W(*this);
    
    return W.destrank();
  }

  // Danger Will Robinson: this rank computation is destructive.
  // if rankcols is given, limit to the first rankcols columns.

int  destrank(int rankcols=-1){

      
    long pcol,prow; // pivot row, column

    long collimit;

    pcol=0;
    prow=0;
    if (rankcols >=0 ){
      collimit=rankcols;
    } else {
      collimit=cols();
    }

  while( pcol < collimit && prow < rows()) {
    int i=prow;
    
    while (i<rows() && (rowlist[i][pcol]==0)){
      i++;
    }
    if (i>=rows()) {
      // rest of this column is zero
      pcol++;
    } else {
      if (i!=prow){
	swap(rowlist[i],rowlist[prow]);
      }

      for (int k=prow+1; k<rows(); k++){
	element_t lead=rowlist[k][pcol];
	if (lead != 0){
	  for (int j=pcol; j< cols(); j++){
	    rowlist[k][j] = rowlist[prow][pcol]*rowlist[k][j] - lead*rowlist[prow][j];
	  }
	}
      }
      pcol++;
      prow++;
    }
  }
  return (std::min(prow,pcol));
}


};



template <typename T> Matrix<T> operator * ( const Matrix<T> & M, const T& scalar){
  Matrix<T> N(M);

  N.mul(scalar);
  return N;
}

template <typename T> Matrix<T> operator * ( const T& scalar, const Matrix<T> & M){
  Matrix<T> N(M);
  N.mul(scalar);
  return N;
}

template <typename T> Matrix<T> operator * ( int i, const Matrix<T> & M){
  return static_cast<const T& >(i)*M;
}


template <typename T>
inline ostream& operator<< (ostream& os, const Matrix<T>& v) {
  typename Matrix<T>::const_iterator j;
  for (j=v.begin() ; j!= v.end(); j++)
    os << *j;
  return os;
};


template <typename T>
inline std::istream& operator >> (std::istream &is, Matrix<T> &v){
  typename Matrix<T>::vecvec::iterator j;
  for (j=v.begin() ; j!= v.end(); j++){
    is >> static_cast< typename  Matrix<T>::vec & >( *j);
  }
  return is;
}

#endif
