// Copyright (C) 2008 David Bremner
//
// Author: David Bremner (bremner@unb.ca)
// Version: $Id$
//
// This file is part of lpmargin.
//
// lpmargin is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// lpmargin is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with lpmargin.  If not, see <http://www.gnu.org/licenses/>.
#include <fstream>
#include <iostream>
#include <algorithm>
#include <cstdlib>
#include <cmath>

#include "power.h"
#include "Symplex.hpp"
#include "QVector.h"
#include "QMatrix.h"
#include "Partition.h"
#include "LP.h"
#include "Random.h"
using namespace std;

enum opt_enum { PRINT_IEQ=1,PRINT_LPSOL=2, LP_VERBOSE=4, PRINT_VERTICES=8, PRINT_EQ=16,PRINT_OBJ=32, PRINT_PARTITION=64, LP_DEBUG=128 };
enum metric_enum { UNKNOWN=0, L_ONE=1, L_INFINITY=1000 };
enum gen_enum {NO_GEN, RANDOM_GEN, EXAUSTIVE_GEN, SYMMETRIC_GEN};

Rational margin(LP & L,const QMatrix & V,const  QMatrix & H, int options){

  int dim=V.cols(), n_points=V.rows();
  
  Partition sigma(n_points);
  Rational smallest_margin(1,1);
  bool first=true;
  QVector e(dim,1);
  QVector zd(dim,0);  


  do {

    if (options&PRINT_PARTITION){
      cout << sigma << endl;
    }

    QMatrix W(2,dim);

    primary_iterator i;
    complement_iterator j;

    W.append(-1*QMatrix::Identity(dim));

    int mink_size=0;
    for ( i=sigma.begin(); i!=sigma.end(); i++){
      for ( j=sigma.cbegin(); j!=sigma.cend(); j++){
	mink_size++;
	W.append(V[*i]-V[*j]);
      }
    }


    
    QMatrix E=W.transpose();
    QVector ze(dim+2,0);
    
    ze[0]=-1;
    ze.append(QVector(mink_size,1));
    
    E.append(ze);

    QVector obj(dim+mink_size+2);

    obj[1]=Rational(1,2);

    if (options & PRINT_OBJ)
      cout << "obj: " << endl << obj;
    if (options & PRINT_EQ)
      cout << "Eq: " << endl << E;
    
    L.set_eq(E);
    L.set_obj(obj);

    QVector zh(H.rows(),0);
    QMatrix Ieq;

    Ieq.append(zh);
    Ieq.append(QVector(H.rows(),1));

    Ieq.append(Rational(-1)*H.transpose());

    for (int i=0; i<mink_size; i++){
      Ieq.append(zh);
    }
    

    Ieq=Ieq.transpose();

    QMatrix NonNeg(mink_size,dim+mink_size+2);
  
    for(int j=0; j<mink_size; j++){
      NonNeg[j][2+dim+j]=1;
    }

    Ieq.append(NonNeg);

    if (options & PRINT_IEQ ) {
      cout << "Ieq: "<< endl << Ieq << endl;
    }
    L.set_ieq(Ieq);
    L.set_eq(E);
    L.solve();

    if (options & PRINT_LPSOL ){
      cout  << "Sol: " << L.obj_value() << ":" << L.opt_vector() << endl;
    }

    if (first || (L.obj_value() < smallest_margin)){
      first=false;
      smallest_margin=L.obj_value();
    }
  } while (smallest_margin>0 && sigma.next());


  return smallest_margin;
}
void gen_points_random(Random& rg,QMatrix &V){

  QVector neg=QVector(V.cols(),-1);
  V[0]=neg;
  for (int i=1; i<V.rows(); i++){
    for (int j=0; j<V.cols(); j++){
      /*  if (j==(i-1) && (i < V.rows()-1))
	V[i][j]=-1;
	else */
	V[i][j]=1-2*(rg.lnext()%2);
    }
  }
  //cerr << V;
}

bool gen_points_symmetric(Symplex & T, QMatrix &V){

  
  for (int i=0; i< V.rows(); i++){
    for (int j=0; j< V.cols(); j++){
      V[i][j]=T[i][j];
    }
  }
  if (!T.next()) 
    return false;
  else
    return true;
}
void gen_points_exaustive(int & index, int iterations, QMatrix &V){
  long mask=index;

  int rank;
  QVector neg=QVector(V.cols(),-1);
  V[0]=neg;

  do {
    
    for (int i=1; i<V.rows(); i++){
      for (int j=0; j<V.cols(); j++){
	//      if (j==(i-1) && (i < V.rows()-1)){
      //	V[i][j]=-1;
      //      } else {
	V[i][j]=(1-2*(mask & 1L))*-1;
	mask >>= 1;
	//}
      }
    }
    rank=V.rank();
    if (rank < min(V.rows(),V.cols())){
      index++;
      mask=index;
    } 
  } while (rank < min(V.rows(), V.cols()) && index < iterations);  


}

int  main(int argc, char **argv){

  int dim=0,n_points=0;
  int optc;
  Rational max_margin=0,margin_target=0;
  int options=0;
  enum gen_enum point_gen=NO_GEN;
  int iterations=0;
  metric_enum metric=UNKNOWN;
  bool find_all=false;
  int loginterval=1000;
  int degen=0;
  Random rg;

  while ((optc=getopt(argc,argv,"i:l:p:d:n:rm:I:sea")) != -1){
    switch(optc){
    case 'i':
      freopen(optarg,"r",stdin);
      break;
    case 'e':
      point_gen=EXAUSTIVE_GEN;
      break;
    case 'a':
      // print all maximal (so far) simplices
      find_all=true;
      break;
    case 'p':
      switch (optarg[0]){
      case 'e':
	options|= PRINT_EQ;
	break;
      case 'i':
	options|= PRINT_IEQ;
	break;
      case 'o':
	options |= PRINT_OBJ;
	break;
      case 'p':
	options |= PRINT_PARTITION;
	break;
      case 's':
	options |= PRINT_LPSOL;
	break;
      case 'v':
	options |= PRINT_VERTICES;
	break;
      case 'l':
	options |= LP_VERBOSE;
	break;
      case 'L':
	options |= LP_VERBOSE;
	options |= LP_DEBUG;
	break;

      default:
	cerr << "illegal verbosity option " << optarg;
      }
      break;
    case 'd':
      dim=atoi(optarg);
      break;
    case 'n':
      n_points=atoi(optarg);
      break;
    case 'I':
      iterations=atoi(optarg);
      break;
    case 'r':
      point_gen=RANDOM_GEN;
      break;
    case 's':
      point_gen=SYMMETRIC_GEN;
      break;
    case 'S':
      rg.set_seeds(optarg);
      break;

    case 't':
      margin_target=atoi(optarg);
      break;
    case 'm':
      switch (optarg[0]){
      case 'i':
	metric=L_INFINITY;
	break;
      case '1':
	metric=L_ONE;
	break;
      default:
	cerr << "Unknown metric " << optarg << endl;
	exit(3);
      }
      break;
    case 'l':
      loginterval=atoi(optarg);
      break;

    default:
      cerr << "Illegal argument" << endl;
      exit(1);
    }
  }



  if (dim == 0 && n_points==0){
    cin >> dim;
    cin >> n_points;
  } else if (dim==0 || n_points==0){
    cerr << "Specify both or neither dimension " 
	 << "and n-points on cmd line." << endl;
    exit(2);
  }

  switch (point_gen){
  case RANDOM_GEN:
    if (iterations==0)
      iterations=100;
    break;
  case EXAUSTIVE_GEN:
    {
      //      int bits=dim*(n_points-1)-(n_points-2);
      int bits=dim*(n_points-1);
      cerr << "bits " << bits << endl;
      if (bits>= 31){
	cerr << "Sorry. I'm exhausted just thinking about that." << endl;
	exit(4);
      };
      iterations= (1L << bits);
    }
		   
    break;
  default:
    iterations=1;
  }
  

 

  QMatrix V(n_points,dim);

  if (point_gen==NO_GEN){
    cin >> V;
  }

  //  cerr << "V: " << endl << V;

  QMatrix H;

  switch(metric){
  case L_INFINITY:
    for (int i=0; i< dim; i++){
      QVector row(dim);
      row[i]=1;
      H.append(row);
      row[i]=-1;
      H.append(row);
    }
    //   cerr << H;
    break;
  case L_ONE:
    
    for (int i=0; i< ddb::power(2,dim); i++){
      QVector row(dim);
      for (int j=0; j<dim; j++){
	int bits=i>>j;
	if (bits & 1) {
	  row[j]=1;
	} else {
	  row[j]=-1;
	}
      }
      H.append(row);
    }
    break;
  default:
    int h_rows;
    cin >> h_rows;
    
    H=QMatrix(h_rows,dim);

    cerr << "Reading unit ball from stdin\n";
    cin >> H;
  }

  //  cerr << "H: " << endl << H;

  LP L;
  
  if (options & LP_VERBOSE)
    L.verbose();

  std::cerr << "options=" <<options << std::endl;
  if (options & LP_DEBUG)
    L.debug();

  QMatrix best_V;

  Symplex T(dim,n_points);

  bool keep_going=true;
  int count=0;
  

  while( keep_going ) {
    
    // generate some points

    Rational new_margin;
    switch (point_gen){
    case RANDOM_GEN:
      gen_points_random(rg,V);
      keep_going=(count>=iterations);
      break;
    case EXAUSTIVE_GEN:
      gen_points_exaustive(count,iterations,V);
      keep_going=(count<=iterations);
      break;
    case SYMMETRIC_GEN:
      keep_going=gen_points_symmetric(T,V);
      break;
    default:
      // do nothing; the points were already read.
      keep_going=false;
      break;
    }

    if (options & PRINT_VERTICES){
      cout << "V:" << V << endl;
    }

    new_margin =  margin(L,V,H,options);
    
    if ((count % loginterval) == 0){
      cout << count+1 <<  "iterations completed " << degen << " degenerate" <<endl;
    }

    if (new_margin > max_margin || (find_all && (new_margin >= max_margin) )){
      max_margin=new_margin;
      best_V=V;
      cout << endl<< "Margin: " <<  max_margin;
      cout << " Achieved by:" << endl <<  V << endl;
    }
    count++;
  }
  cout << "Max Margin " << max_margin << endl;
  cout << "Vertices: " << endl << best_V << endl;
  cout << iterations <<  "total iterations completed " << degen << " degenerate" <<endl;

  cout << "Random State:" << rg << endl;

}


